@@ -1,13 +1,15 @@
 /*
  * feature.c --- convert between features and strings
- * 
+ *
  * Copyright (C) 1999  Theodore Ts'o <tytso@mit.edu>
- * 
- * This file can be redistributed under the terms of the GNU Library General
- * Public License
- * 
+ *
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
+ * %End-Header%
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -32,8 +34,6 @@ const char *e2p_hash2string(int num)
 {
 	struct hash  *p;
 	static char buf[20];
-	char	fchar;
-	int	fnum;
 
 	for (p = hash_list; p->string; p++) {
 		if (num == p->num)
